#!/usr/bin/env node
/**
 * Remote Light Control using Blynk
 * Filename: remote-light.js
 * Tasks:
 * * Check network interfaces using shell command ifconfig
 * * Print interface statuses to OLED screen
 * * Connect to Blynk service and make syncronization of states
 * * Communicate with Blynk server, make actions
 * @author  Mario Simunic
 * @email mario.simunic@gmail.com
 * @version 1.0.60, 06/14/2018
 */
 
/* You should get Auth Token in the Blynk App.
 * Go to the Project Settings (nut icon), or check your email.
 */
const auth = "a232f4e00180481fbb7560f9bfc390fc";

/* RegExp for integers extraction */
const numberPattern = "/\d+/g"; 

/* PACKAGES */
const sys = require('sys');		/*!< Require for non-blocking (asynchronous) system functions ie. ReadFile() */
const oledExp = require("/usr/bin/node-oled-exp");	/*!< Require for printing text to OLED screen */
const BlynkLib = require('/usr/bin/blynk-library');	/*!< Require for communication with Blynk */
var fs = require('fs');
var exec = require('child_process').exec;	/*!< Require for non-blocking (asynchronous) system functions ie. exec() */
var Gpio = require('/usr/bin/onoff-node/onoff.js').Gpio; 	/*!< Require for access to GPIO pins */

/* GLOBAL VARIABLES */
var child, hostname, text;
var apcli0ConnectionStatus;
var timeString;
var alarmStartSeconds;
var alarmStopSeconds;
var alarmDays, alarmDaysLength;
var alarmEvent, alarmState;
var alarmEnabled;

var lightState, lightDuty;
var lightChanged;

var alarmActivationTime;
var alarmActionDelay;		// seconds
var alarmDelayEnabled;

var inputAState, inputBState;
var inputAChange, inputBChange;

var blinkingTaskID;
var alarmTaskID;
var alarmTaskEnabled;
var iftttTaskID;
var timeCnt;

var presenceState, inBedState, alarmActionEvent;
var blinkingEnabled, timeoutEnable;
var inBedTriggerEnabled, presenceTriggerEnabled;

var digpotValue;	// present value of digital potentiometer

/* PHERIPHERALS*/
var gpioLight = new Gpio(0, 'out');
var CS = new Gpio(6, 'high');	//!< 'high' = output with logic "1" as initial state
var UD = new Gpio(18, 'high');
var INC = new Gpio(19, 'high');	//!< active on falling edge
var gpioInputA = new Gpio(3, 'in', 'rising',  {debounceTimeout: 10}); 	// PIR sensor
var gpioInputB = new Gpio(2, 'in', 'both',  {debounceTimeout: 10});		// Something else

/* BLYNK */
// Blynk
var blynk = new BlynkLib.Blynk(auth);
var buttonLight = new blynk.VirtualPin(0);		//!< Blynk Button for Light
var sliderLight = new blynk.VirtualPin(1); 		//!< Blynk Slider for Light Dimming
//var terminal = new blynk.VirtualPin(3); 		//!< Terminal widget
var timeInput = new blynk.VirtualPin(5); 		//!< Blynk Time Input

/* GPIO inputs*/
gpioInputA.watch(function (err, value) {
	inputAState = value;
	inputAChange = 1;
});

gpioInputB.watch(function (err, value) {
	inputBState = value;
	inputBChange = 1;
});

function digpotSet(value){
	var i, diff;
	if (value <= 100 && value >= 0 ){
		if (value < digpotValue){
			UD.writeSync(0);
			diff = digpotValue - value;
		} else{
			UD.writeSync(1);
			diff = value - digpotValue;
		}
		CS.writeSync(0);
		for (i = 0; i < diff+1; i++){
				INC.writeSync(0);
				a = setTimeout(function(){},1);
				INC.writeSync(1);
				clearTimeout(a);
			}
		CS.writeSync(1);
		digpotValue = value;
	}
	UD.writeSync(1);
}

function digpotInit(){
	var i, a;
	CS.writeSync(0);	// select digpot
	for(i=0;i<101;i++){
		INC.writeSync(0);
		a = setTimeout(function(){},1);
		INC.writeSync(1);
		clearTimeout(a);
	}
	CS.writeSync(1);	// select digpot
	digpotValue = i-1;
}

/** 
 * Reads host name
 */
function readHostname(){
	var result = require('child_process').execSync('uci get system.@system[0].hostname').toString();
	return result;
}

/**
 * Reads status of Omega's acces-point
 */
function readStatusWlan(){
	var result = require('child_process').execSync('uci get wireless.ap.ifname').toString().trim();
	var tmp = result;
	result = require('child_process').execSync('uci get wireless.ap.ssid').toString().trim();
	return (tmp + ": " + result);
}

/**
 * Reads IP Address of Omega's acces-point
 */
function readAddressWlan(){
	var result = require('child_process').execSync('ifconfig | grep -i wlan -A 1 | grep -i inet').toString().trim();
	result = result.substr(result.indexOf(':') + 1);
	var addr = result.substr(0,result.indexOf(' '));
	return (" " + addr);
}

/** 
 * Reads SSID of wlan to which Omega is connected
 */
function getSSIDApcli0(){
    var text = require('child_process').execSync('iwconfig apcli0 | grep -i ssid').toString().trim();
    var start = text.indexOf('ESSID:"');
    text = text.substr(start+7);
    var stop = text.indexOf('"');
    text = text.substr(0,stop);
//  fs.writeFileSync('/root/ssid-connected.log', stop.toString());
	if ( stop === 0 ){
		apcli0ConnectionStatus = 0 ;
	} else {
		apcli0ConnectionStatus = 1;
	}
return ("apcli0: " + text);
}

/**
 * Reads IP Address of wlan to which Omega is connected
 */
function readAddressApcli0(){
	var result = require('child_process').execSync('ifconfig | grep -i apcli0 -A 1 | grep -i inet').toString().trim();
	result = result.substr(result.indexOf(':') + 1);
	var addr = result.substr(0,result.indexOf(' '));
	return (" " + addr);
}

/**
 * Initialize OLED screen
 */
function oledInit() {
  // Initialize OLED screen
  oledExp.init();
  oledExp.setDisplayPower(1);
  oledExp.setBrightness(127);
  oledExp.setDim(0);
  oledExp.clear();
}

/**
 * Print message to OLED screen
 * @param[in]  line  The line number at which begining print message
 * @param[in]  message  Message string to print
 */
function oledPrint(line, message){
    oledExp.setCursor(line,0); // Position to the begining of the line
    oledExp.write("                     "); // clear the line
    oledExp.setCursor(line,0); // Position at beginning of the line
    oledExp.write(message);  // Print message
}

/**
 * Get Date and Time and return string for peinting to OLED screen
 */
function getDateTime() {
    var date = new Date();
    var hour = date.getHours();
    hour = (hour < 10 ? "0" : "") + hour;
    var min  = date.getMinutes();
    min = (min < 10 ? "0" : "") + min;
    var sec  = date.getSeconds();
    sec = (sec < 10 ? "0" : "") + sec;
    var year = date.getFullYear();
    var month = date.getMonth() + 1;
    month = (month < 10 ? "0" : "") + month;
    var day  = date.getDate();
    day = (day < 10 ? "0" : "") + day;
   // var time = date.getTime(); // Milisecond counter
   // return year + "-" + month + "-" + day + "  " + hour + ":" + min + ":" + sec;
   return year + "-" + month + "-" + day + "  " + hour + ":" + min;
}

/**
 * Set GPIO which controls the light
 * Update all virtual controllers
 * @param [in] Light state tu be pushed to App Button
 * @param [in] Light Duty tu be pushed to App Slider
 */
function lightUpdate(state, duty){
	if (state === 1){
		gpioLight.writeSync(1);
	} else {
		gpioLight.writeSync(0);
	}
	virtualSliderUpdate(duty);
	virtualButtonUpdate(state);
	lightChanged = 0;
}

/* ----------------------------------------------------------------------
 * BLYNK FUNCTIONS
 */
 
/*!
 * Catch errors
 */
blynk.on('error', (err) => {
	// What to do with errors?
	// Writing to console breaks node.js script execution
	// --> Do nothing.
	//console.error('whoops! there was an error');
});

/**
 * Handle Blynk Light Button Change
 */
buttonLight.on('write', function(param){
	lightState = parseInt(param);
	lightChanged = 1;
	if (lightState === 1){
		lightDuty = 100;
	} else {
		lightDuty = 0;
	}
	alarmEnabled = 0;
	alarmEvent = 0;
	lightUpdate(lightState, lightDuty);
});

/**
 * Handle Blynk Light Slider Change
 */
sliderLight.on('write', function(param){
	lightDuty = parseInt(param);
	lightChanged = 1;
	if (lightDuty > 0){
		lightState = 1;
	} else {
		lightState = 0;
	}
	alarmEnabled = 0;
	alarmEvent = 0;
	lightUpdate(lightState, lightDuty);
});
 
/**
 * Handle Blynk new alarm time input
 */
timeInput.on('write', function(param){
	var i;
	timeString = param.toString().trim();
	alarmEnabled = 1;
	presenceState = 0;		//!< Reset presence to ease demostration
	parseAlarmTime(timeString);
});

/**
 * Update state of mobile App BUTTON for light
 * @param [in] New state to be pushed to Mobile App
 */
function virtualButtonUpdate(state){
	if (state === 0){ 					//!< Set button (in Blynk App) color for state OFF
		blynk.setProperty(0, "color", "#0BE29E");	//!< Update Button Color to Gray (virtual pin 0)
	} else {       						//!< Set button (in Blynk App) color for state ON
		blynk.setProperty(0, "color", "#9AA7AF");	//!< Update Button Color to Green (virtual pin 0)
	}
	
	blynk.virtualWrite(0,state);					//!< Update in App Button State
}

/**
 * Update state of mobile App SLIDER for light dimming
 * @param [in] New state to be pushed to Mobile App
 */
function virtualSliderUpdate(duty){
	if (duty === 0){ 					//!< Set slider (in Blynk App) color for state OFF
		blynk.setProperty(1, "color", "#0BE29E");	//!< Update Slider Color to Gray (virtual pin 1)
	} else {       						//!< Set slider (in Blynk App) color for state ON
		blynk.setProperty(1, "color", "#9AA7AF");	//!< Update Slider Color to Green (virtual pin 1)
	}
	
	blynk.virtualWrite(1,duty);					//!< Update in App Slider Value
}

/**
 * Parse time from Time Input Widget param string
 /* timeString = 'StartTime,StopTime,TimeZone,Day1 - Day7,TimeZoneOffset'	
 * ie. * timeString = 7140,34500,Europe/Zagreb,1,7,7200
 * StartTime, StopTime and TimeZoneOffset - in seconds from midnight (0:00 AM)
 * ie. 1:00 AM = 1*60min*60sec = 3600
 * ie. 1:05 AM = (1*60min + 5min)*60sec = 3900
 * TimeZoneOffset - in seconds from UTC
 * Days:
 * Monday = 1
 * Tuesday = 2
 * Wednesday = 3
 * Thursday = 4
 * Friday = 5
 * Saturday = 6
 * Sunday = 7
 */
function parseAlarmTime(timestr){
	var index;
	timestr = timestr.toString();
	alarmStartSeconds = parseInt(timestr);				//!< First Integer - StartTime
	timestr = timestr.substr(timestr.indexOf(',') + 1); //!< Cut timestr
	alarmStopSeconds = parseInt(timestr); 				//!< Second Integer - Stop Time
	timestr = timestr.substr(timestr.indexOf(',') + 1);	//!< Cut timestr
	timestr = timestr.substr(timestr.indexOf(',') + 1);	//!< Cut time zone
	/** Parse days active*/
	for (index = 0; index < 7; index++){
		alarmDays[index] = 0;
	}
	index = 0;
	while((alarmDays[index] = parseInt(timestr)) < 8){
		timestr = timestr.substr(timestr.indexOf(',') + 1);
		index = index + 1;
	}
	alarmDaysLength = index;
}

/**
 * Check if it's time for alarm
 * return 0: - alarm disabled
 * return 1: - alarm wasn't active, inactive
 * return 2: - alarm active
 * return 3: - alarm was active, inactive
 */
function checkAlarm(){
	var date = new Date();
	var todaySeconds = (date.getHours() * 60 + date.getMinutes())*60 + date.getSeconds();
	var day = date.getDay();		//!< getDay() returns weekday number [0 - 6], 0 = Sunday, 1 = Monday
	var i = 0, alarmToday = 0;
	
	if (day === 0){
		day = 7;	//!< adjust date.getDay() format to TimeInput format
	}
	// Check weather alarm should be activated today
	for (i=0;i<alarmDaysLength;i++){
		if (alarmDays[i] === day){
			alarmToday = 1;
		}
	}
	
	if (alarmToday === 1){		
		if (alarmStopSeconds >= alarmStartSeconds){	// Uobicajno ponasanje
			if (todaySeconds > alarmStartSeconds){ 	// vrijeme za aktivaciju
				alarmState = 2;							// Aktivan
			}
			if (todaySeconds > alarmStopSeconds){	// vrijeme za deaktivaciju
				alarmState = 3;							// dekativiran, bio je aktivan
			}
		}
		
		if (alarmStartSeconds > alarmStopSeconds){ 	// alarm koji kasno starta i rano sljedeci dan mora se zaustaviti
			if (todaySeconds > alarmStopSeconds){	// vrijeme za deaktivaciju
				alarmState = 3;							// deaktiviran, bio je aktivan
			}
			if (todaySeconds > alarmStartSeconds){	// vrijeme za start
				alarmState = 2;							// aktivan
			}
		}
		
		if (alarmState === 3 && alarmEvent === 0) {		// ako je bio aktivan i deaktiviran je
			alarmState = 1;								
			alarmEnabled = 1;							//!< Re-Enable alarm
		}
	} else{
		alarmState = 0;
	}
	
	return alarmState;
}


/**
 * Start periodic blinking action
 */
function runBlynking() {
	var num = 1;
	/* problem: setInterval() return an object which changes elements depending on weather is it started or killed */
	if (parseInt(blinkingTaskID) == 0){		//!< This is not exactly JS way of doing it, but it works
		blinkingTaskID = setInterval((function(){
			this.num ^= 1;
			// gpioLight.writeSync(this.num);
			lightState = this.num;
			lightDuty = this.num * 100;
			lightUpdate(lightState, lightDuty);
		}), 667);
	}
}

/**
 * Kill periodic blinking action
 */
function stopBlinking() {
	clearInterval(blinkingTaskID);
	blinkingTaskID = 0;		//!< To be abble to check is setInterval active
}

function alarmTask(){
	if (alarmEvent === 1){ 	//!< If alarm is active
		alarmDelayEnabled = 1;
		if (blinkingEnabled === 1){
			runBlynking();
		} else {
			stopBlinking();
			lightState = 1;
			lightDuty = 100;
			lightUpdate(lightState, lightDuty);
		}
	} else {
		stopBlinking();
		alarmDelayEnabled = 0;
		blinkingEnabled = 0;
// Debugging
//		lightState = 0;
//		lightDuty = 0;
//		lightUpdate(lightState, lightDuty);
	}
}

/**
 * Simple function to init intialize all global variables (states)
 * Run once at start
 */
function initStates(){
	// Initilize variables
	/* Read hostname (only once, at startaup) */
	hostname = readHostname();
	
	lightState = 0; 
	lightDuty = 0;
	
	// Call init functions
	lightUpdate(lightState, lightDuty);
	
	/* Call OLED initialization */
	oledInit();
	
	/* Initialize digital potentiometer */
	digpotInit()
	
	/* blinking function at alarm enabled*/
	blinkingEnabled = 0;

	/* timeout wait enebaled */
	timeoutEnable = 1;
	
	apcli0ConnectionStatus = 0;
	timeString = "";
	
	alarmStartSeconds = 0;
	alarmStopSeconds = 0;
	alarmDays = []; 
	alarmDaysLength = 0;
	alarmEvent = 0; 
	alarmState = 0;
	alarmEnabled = 1;

	lightChanged = 0;
	
	
	presenceState = 0; 
	inBedState = 0; 
	
	alarmActionDelay = 10;		// seconds
	alarmDelayEnabled = 0;
	alarmActionEvent = 0;
	
	// interval tasks ID
	blinkingTaskID = 0;
	alarmTaskID = 0;
	iftttTaskID = 0;
	
	// IFTTT tasks
	alarmTaskEnabled = 1;		// IF Alarm THEN ...
	inBedTriggerEnabled = 0; 
	presenceTriggerEnabled = 1;	// IF movement detected THEN ...
	
	timeCnt = 0;
}

/**
 * My small If This Than That function.
 * Implemet behaviour for light based on events
 */
function myIFTTT(){
	
	if (alarmTaskEnabled === 1){
		if (parseInt(alarmTaskID) == 0){		//!< This is not exactly JS way of doing it, but it works
			alarmTaskID = setInterval(alarmTask, 1000);	//!< Action to do when alarm active
		}
	} else {
		clearInterval(alarmTaskID);
		alarmTaskID = 0;
	}
	
	/**
	 * inBedTrigger Action
	 */
	if (inBedTriggerEnabled === 1){
		//Action to do on inBed trigger
		if (inBedState === 1 && alarmActionEvent == 1){
			// Action to do when inBed
			blinkingEnabled = 1;
		} else{
			// Action to do when not inBed
			alarmActionEvent = 0;
			blinkingEnabled = 0;
		}
	} else{
		// Action to do when there is no inBed trigger
	}
	
	/**
	 * presence Action
	 */
	if (presenceTriggerEnabled === 1){
		//Action to do on presence trigger
		if (alarmEvent === 1){		//!< if ongoing active alarm
			if (presenceState === 1){	//!< if presence detected
				// Action to do when somebody is in a Room
				inBedTriggerEnabled = 0;	//!< presence disables inBed sensor
				blinkingEnabled = 0;		//!< stop blinking
				stopBlinking();				//!< kill blinking function
			} else{
				// Action to do when not present in a Room
			}
		} else {
			// Action to do if presenceTrigger Enabled, but alarm isn't active
		}
	} else{
		// Action to do when there is no inBed trigger
	}
	
	/* Alarm Action Delay Event */
	if (alarmDelayEnabled === 1){
		timeCnt = timeCnt +1 ;
		if ( timeCnt > alarmActionDelay ){
			// alarmActionEvent = 1;
			blinkingEnabled = 1;
			timeCnt = 0;
		}
	}
}

/* --------------------------------------------------- */
/* CALL FUNCTIONS */

/* Initialize variables */
initStates();

/* RUN IFTTT */
iftttTaskID = setInterval(myIFTTT, 1000);

/** 
 * Update OLED screen time
 * Run every 1 seconds. If full minute (00 seconds) update time.
 * Every 5 minute clear OLED
 */
var num = 1;
var cnt = 0;
var countMinutes = 0;
setInterval(function() {
	var date = new Date();
	var time = getDateTime();
	/* Update OLED */
	if ( date.getSeconds() === 0 ) { 	//!< Every 1 minute
		countMinutes = countMinutes + 1;//!< Count minutes
		if (countMinutes === 5) {		//!< Every 5 minutes
			oledExp.clear();			//!< Clear screen before printing the time
			countMinutes = 0;			//!< reset counter
			
			presenceState = 0;			//!< Clear presence			
		}
		oledPrint(0, time); // Print date and time
	}
	
	/* Service Alarm */
	if (alarmEnabled === 1 && alarmEvent === 0 && checkAlarm() === 2 ){ 		// if alarm is active
		alarmEvent = 1;
		alarmActivationTime = date.getSeconds();	// remember time of activation
	} 
	if (alarmEnabled === 1 && alarmEvent === 1 && checkAlarm() === 3 ){
		alarmEvent = 0;
	}
	
	/* Service GPIO inputs */
	/** InputA Change - presence in a room */
	if ( inputAChange === 1 && inputAState === 1){
		presenceState = 1;
		inputAChange = 0;
	}
	
	/** InputB Change Event - presence in a Bed */
	if ( inputBChange === 1 ){
		inBedState= inputBState;
		inputBChange = 0;
	}
	
}, 1000);	// Every 1 seconds

/** 
 * Every 23s (=23000 ms) update OLED screen time
 */
setInterval(function() {
	oledPrint(1, hostname);
	text = readStatusWlan();
	oledPrint(2, text);
	text = readAddressWlan();
	oledPrint(3, text);
	text = getSSIDApcli0();
	oledPrint(4, text);
	text = readAddressApcli0();
	oledPrint(5, text);
	
}, 123667); // Every 123.667 seconds
